from importlib.resources import files

TEST_OUTPUT_DIR = files("tests").joinpath("sample_output")

BINARY_CLASSIFICATION_SMALL_PATH = files("tests").joinpath(
    "sample_predictions/binary_classification/small.parquet"
)



"""Main entry point for the meds_evaluation package."""
import json
from datetime import datetime
from importlib.resources import files
from pathlib import Path

import hydra
import polars as pl
from loguru import logger
from omegaconf import DictConfig, OmegaConf

from meds_evaluation.evaluate import evaluate_binary_classification

config_yaml = files("meds_evaluation").joinpath("configs/meds_evaluation.yaml")


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig) -> None:
    st = datetime.now()
    logger.info(f"Starting evaluation with config:\n{OmegaConf.to_yaml(cfg)}")

    # Load the model prediction dataframe
    logger.info(f"Loading model predictions dataframe '{cfg.predictions_path}'")
    predictions = pl.read_parquet(cfg.predictions_path)

    # Set output path
    evaluation_output_dir = Path(cfg.output_dir)
    evaluation_output_dir.mkdir(exist_ok=True, parents=True)

    # Run the evaluation
    logger.info("Running evaluation...")
    result = evaluate_binary_classification(
        predictions, samples_per_subject=cfg.samples_per_subject, resampling_seed=cfg.resampling_seed
    )

    # Save the results
    with open(evaluation_output_dir / "results.json", "w") as f:
        json.dump(result, f, indent=4)

    logger.info(
        f"Completed in {datetime.now() - st}. Results saved to '"
        f"{evaluation_output_dir / 'results.json'}'."
    )


if __name__ == "__main__":
    main()



import polars as pl

# Required fields
SUBJECT_ID_FIELD = "subject_id"
PREDICTION_TIME_FIELD = "prediction_time"

# Ground truth field
BOOLEAN_VALUE_FIELD = "boolean_value"

# Prediction fields
PREDICTED_BOOLEAN_VALUE_FIELD = "predicted_boolean_value"
PREDICTED_BOOLEAN_PROBABILITY_FIELD = "predicted_boolean_probability"


REQUIRED_FIELDS = {SUBJECT_ID_FIELD, PREDICTION_TIME_FIELD, BOOLEAN_VALUE_FIELD}
PREDICTION_FIELDS = {PREDICTED_BOOLEAN_VALUE_FIELD, PREDICTED_BOOLEAN_PROBABILITY_FIELD}

BINARY_CLASSIFICATION_SCHEMA_DICT = {
    "subject_id": pl.Int64,
    "prediction_time": pl.Datetime,
    "boolean_value": pl.Boolean,
    "predicted_boolean_value": pl.Boolean,
    "predicted_boolean_probability": pl.Float64,
}


def validate_binary_classification_schema(df: pl.DataFrame) -> None:
    """Checks if the predictions dataframe contains the necessary columns for binary classification metrics.

    Args:
        predictions: a DataFrame following the MEDS prediction schema for binary classification, containing at
        least one of the binary classification columns: "predicted_boolean_value",
        "predicted_boolean_probability".

    Raises:
        ValueError: if the predictions dataframe does not contain the necessary columns.
    """
    df_type_dict = dict(df.schema)

    # Check required fields
    df_fields = set(df_type_dict.keys())
    missing_required_fields = REQUIRED_FIELDS - df_fields
    if missing_required_fields:
        raise ValueError(f"Missing required fields: {missing_required_fields}")
    else:
        for required_field in REQUIRED_FIELDS:
            if df_type_dict[required_field] != BINARY_CLASSIFICATION_SCHEMA_DICT[required_field]:
                raise ValueError(
                    f"Mismatched type for {required_field}: expected {df_type_dict[required_field]}, "
                    f"got {BINARY_CLASSIFICATION_SCHEMA_DICT[required_field]}"
                )

    # Check at least one of the prediction fields is present
    prediction_fields = PREDICTION_FIELDS & df_fields
    if not prediction_fields:
        raise ValueError(f"Missing all prediction fields: {PREDICTION_FIELDS}")
    elif all(df[field].is_null().all() for field in prediction_fields):
        raise ValueError(
            f"At least one of the prediction fields should have non-null values: {PREDICTION_FIELDS}"
        )
    else:
        for prediction_field in prediction_fields:
            if df_type_dict[prediction_field] != BINARY_CLASSIFICATION_SCHEMA_DICT[prediction_field]:
                raise ValueError(
                    f"Mismatched type for {prediction_field}: expected {df_type_dict[prediction_field]}, "
                    f"got {BINARY_CLASSIFICATION_SCHEMA_DICT[prediction_field]}"
                )



"""Methods for evaluating different types of tasks and subpopulations on a standard set of metrics.

Most metrics will be directly based on sklearn implementation, and a standard set will be defined for
binary metrics initially. (TODO: add multiclass, multilabel, regression, ... evaluation).

See
    https://scikit-learn.org/stable/api/sklearn.metrics.html
    https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics

Additionally, functionality for evaluating metrics on a per-sample vs per-subject basis is provided to
ensure balanced representation of all subjects in the dataset.

TODO fairness functionality and filtering populations based on complex user-defined criteria.
"""

import numpy as np
import polars as pl
from numpy.typing import ArrayLike
from sklearn.calibration import calibration_curve
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    f1_score,
    precision_recall_curve,
    roc_auc_score,
    roc_curve,
)

from meds_evaluation.schema import (
    BOOLEAN_VALUE_FIELD,
    PREDICTED_BOOLEAN_PROBABILITY_FIELD,
    PREDICTED_BOOLEAN_VALUE_FIELD,
    SUBJECT_ID_FIELD,
    validate_binary_classification_schema,
)
from meds_evaluation.utils import _resample

# TODO: input processing for different types of tasks
#   detect which set of metrics to obtain based on the task and the contents of the model prediction dataframe


def evaluate_binary_classification(
    predictions: pl.DataFrame, samples_per_subject=4, resampling_seed=0
) -> dict[str, dict[str, float | list[ArrayLike]]]:
    """Evaluates a set of model predictions for binary classification tasks.

    Args:
        predictions: a DataFrame following the MEDS label schema and additional columns for
        "predicted_value" and "predicted_probability".
        samples_per_subject: the number of samples to take for each unique subject_id in the dataframe for
        per-subject metrics.
        resampling_seed: random seed for resampling the dataframe.
        # TODO consider adding a parameter for the metric set to evaluate

    Returns:
        A dictionary mapping the metric names to their values.
        The visual (curve-based) metrics will return the raw values needed to create the plot.

    Raises:
        ValueError: if the predictions dataframe does not contain the necessary columns.
    """
    # Verify the dataframe schema to contain required fields for the binary classification metrics
    validate_binary_classification_schema(predictions)

    true_values = predictions[BOOLEAN_VALUE_FIELD]

    predicted_values = predictions[PREDICTED_BOOLEAN_VALUE_FIELD]
    predicted_probabilities = predictions[PREDICTED_BOOLEAN_PROBABILITY_FIELD]

    resampled_predictions = _resample(
        predictions,
        sampling_column=SUBJECT_ID_FIELD,
        n_samples=samples_per_subject,
        random_seed=resampling_seed,
    )

    true_values_resampled = resampled_predictions[BOOLEAN_VALUE_FIELD]
    predicted_values_resampled = resampled_predictions[PREDICTED_BOOLEAN_VALUE_FIELD]
    predicted_probabilities_resampled = resampled_predictions[PREDICTED_BOOLEAN_PROBABILITY_FIELD]

    if predicted_values.is_null().all():
        predicted_values = None
        predicted_values_resampled = None

    if predicted_probabilities.is_null().all():
        predicted_probabilities = None
        predicted_probabilities_resampled = None

    results = {
        "samples_equally_weighted": _get_binary_classification_metrics(
            true_values, predicted_values, predicted_probabilities
        ),
        "subjects_equally_weighted": _get_binary_classification_metrics(
            true_values_resampled, predicted_values_resampled, predicted_probabilities_resampled
        ),
    }

    # TODO write to output file
    return results


def _get_binary_classification_metrics(
    true_values: ArrayLike,
    predicted_values: ArrayLike | None,
    predicted_probabilities: ArrayLike | None,
) -> dict[str, float | list[ArrayLike]]:
    """Calculates a set of binary classification metrics based on the true and predicted values.

    Args:
        true_values: the true binary values
        predicted_values: the predicted binary values
        predicted_probabilities: the predicted probabilities
        TODO consider the list of metrics

    Returns:
        A dictionary mapping the metric names to their values.
        The visual (curve-based) metrics will return the raw values needed to create the plot.
    """
    results = {}

    if predicted_values is not None:
        results["binary_accuracy"] = accuracy_score(true_values, predicted_values)
        results["f1_score"] = f1_score(true_values, predicted_values).item()

    if predicted_probabilities is not None:
        results["roc_auc_score"] = roc_auc_score(true_values, predicted_probabilities).item()
        results["average_precision_score"] = average_precision_score(
            true_values, predicted_probabilities
        ).item()

        r = roc_curve(true_values, predicted_probabilities)
        results["roc_curve"] = [r[0].tolist(), r[1].tolist()]

        p = precision_recall_curve(true_values, predicted_probabilities)
        results["precision_recall_curve"] = [p[0].tolist(), p[1].tolist()]

        c = calibration_curve(true_values, predicted_probabilities, n_bins=10)
        results["calibration_curve"] = [c[0].tolist(), c[1].tolist()]
        results["calibration_error"] = np.abs(c[0] - c[1]).mean().item()

    return results



import numpy as np
import polars as pl

from meds_evaluation.schema import SUBJECT_ID_FIELD


def _resample(
    predictions: pl.DataFrame, sampling_column=SUBJECT_ID_FIELD, n_samples=1, random_seed=0
) -> pl.DataFrame:
    """Samples (with replacement) the dataframe to represent each unique value in the sampling column equally.

    Args:
        predictions: a dataframe following the MEDS label schema
        sampling_column: the dataframe column according to which to resample
        n_samples: the number of samples to take for each unique value in the sample_by column

    Returns:
        A resampled dataframe with n_samples for each unique value in the sample_by column.

    Raises:
        ValueError: if the sampling column is not present in the predictions dataframe

    Examples:
    >>> _resample(pl.DataFrame({"a": [1, 2, 3, 4, 5, 6]}))
    Traceback (most recent call last):
    ...
    ValueError: The model prediction dataframe does not contain the "subject_id" column.
    >>> _resample(pl.DataFrame({"subject_id": [1, 2, 2, 3, 3, 3]}))
    shape: (3, 1)
    ┌────────────┐
    │ subject_id │
    │ ---        │
    │ i64        │
    ╞════════════╡
    │ 1          │
    │ 2          │
    │ 3          │
    └────────────┘
    >>> _resample(pl.DataFrame({"subject_id": [1, 2, 2, 3, 3, 3]}), n_samples=2)
    shape: (6, 1)
    ┌────────────┐
    │ subject_id │
    │ ---        │
    │ i64        │
    ╞════════════╡
    │ 1          │
    │ 1          │
    │ 2          │
    │ 2          │
    │ 3          │
    │ 3          │
    └────────────┘
    >>> _resample(pl.DataFrame({"subject_id": [1, 3, 4, 2, 2, 3, 3, 3, 1]}), n_samples=1)
    shape: (4, 1)
    ┌────────────┐
    │ subject_id │
    │ ---        │
    │ i64        │
    ╞════════════╡
    │ 1          │
    │ 2          │
    │ 3          │
    │ 4          │
    └────────────┘
    >>> _resample(pl.DataFrame({"a": [1, 3, 4, 2, 2, 3, 3, 3, 1]}), sampling_column="a", n_samples=1)
    shape: (4, 1)
    ┌─────┐
    │ a   │
    │ --- │
    │ i64 │
    ╞═════╡
    │ 1   │
    │ 2   │
    │ 3   │
    │ 4   │
    └─────┘
    """

    # TODO resampling empty dataframe should throw an error

    if sampling_column not in predictions.columns:
        raise ValueError(f'The model prediction dataframe does not contain the "{sampling_column}" column.')

    predictions_sorted = predictions.sort(sampling_column)
    sampling_column = predictions_sorted[sampling_column].to_numpy()

    # Split the indices of the dataframe by the unique values in the sampling column
    np.random.seed(random_seed)
    splits = np.split(np.arange(len(sampling_column)), np.unique(sampling_column, return_index=True)[1][1:])
    resampled_ids = np.concatenate(
        [np.random.choice(split, n_samples, replace=True) for split in splits]
    ).tolist()

    return predictions_sorted[resampled_ids]



"""Integration test for the binary classification evaluation process."""
import json
from datetime import datetime
from pathlib import Path

import polars as pl
import pytest
from omegaconf import DictConfig

from meds_evaluation.__main__ import main as binary_classification_main
from meds_evaluation.evaluate import evaluate_binary_classification
from meds_evaluation.schema import BINARY_CLASSIFICATION_SCHEMA_DICT
from tests import BINARY_CLASSIFICATION_SMALL_PATH, TEST_OUTPUT_DIR

SAMPLE_CONFIG = DictConfig(
    {
        "predictions_path": BINARY_CLASSIFICATION_SMALL_PATH,
        "output_dir": TEST_OUTPUT_DIR,
        "samples_per_subject": 4,
        "resampling_seed": 0,
    }
)


BINARY_CLASSIFICATION_SMALL = pl.DataFrame(
    {
        "subject_id": [1, 2, 3],
        "prediction_time": [
            datetime(2020, 1, 1, 12, 0, 0),
            datetime(2020, 1, 1, 12, 0, 0),
            datetime(2020, 1, 1, 12, 0, 0),
        ],
        "boolean_value": [True, False, True],
        "predicted_boolean_value": [True, False, False],
        "predicted_boolean_probability": [0.9, 0.1, 0.2],
    },
    schema=BINARY_CLASSIFICATION_SCHEMA_DICT,
)

EXPECTED_OUTPUT_SMALL = {
    "samples_equally_weighted": pytest.approx(
        {
            "binary_accuracy": 0.6666666666666666,
            "f1_score": 0.6666666666666666,
            "roc_auc_score": 1.0,
            "average_precision_score": 1.0,
            "roc_curve": [[0.0, 0.0, 0.0, 1.0], [0.0, 0.5, 1.0, 1.0]],
            "precision_recall_curve": [[0.6666666666666666, 1.0, 1.0, 1.0], [1.0, 1.0, 0.5, 0.0]],
            "calibration_curve": [[0.0, 1.0, 1.0], [0.1, 0.2, 0.9]],
            "calibration_error": 0.3333333333333333,
        }
    ),
    "subjects_equally_weighted": pytest.approx(
        {
            "binary_accuracy": 0.6666666666666666,
            "f1_score": 0.6666666666666666,
            "roc_auc_score": 1.0,
            "average_precision_score": 1.0,
            "roc_curve": [[0.0, 0.0, 0.0, 1.0], [0.0, 0.5, 1.0, 1.0]],
            "precision_recall_curve": [[0.6666666666666666, 1.0, 1.0, 1.0], [1.0, 1.0, 0.5, 0.0]],
            "calibration_curve": [[0.0, 1.0, 1.0], [0.1, 0.2, 0.9]],
            "calibration_error": 0.3333333333333333,
        }
    ),
}


def test_main_binary_classification():
    binary_classification_main(SAMPLE_CONFIG)
    # TODO potentially use temporary files
    with open(Path(SAMPLE_CONFIG.output_dir) / "results.json") as f:
        actual_output = json.load(f)

    assert actual_output == EXPECTED_OUTPUT_SMALL


def test_evaluate_binary_classification_small():
    actual_output = evaluate_binary_classification(BINARY_CLASSIFICATION_SMALL)
    assert actual_output == EXPECTED_OUTPUT_SMALL


def test_evaluate_binary_classification_small_null_values():
    input = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
            "predicted_boolean_value": [None, None, None],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
        },
        schema=BINARY_CLASSIFICATION_SCHEMA_DICT,
    )

    expected_output = {
        "samples_equally_weighted": pytest.approx(
            {
                "roc_auc_score": 1.0,
                "average_precision_score": 1.0,
                "roc_curve": [[0.0, 0.0, 0.0, 1.0], [0.0, 0.5, 1.0, 1.0]],
                "precision_recall_curve": [[0.6666666666666666, 1.0, 1.0, 1.0], [1.0, 1.0, 0.5, 0.0]],
                "calibration_curve": [[0.0, 1.0, 1.0], [0.1, 0.2, 0.9]],
                "calibration_error": 0.3333333333333333,
            }
        ),
        "subjects_equally_weighted": pytest.approx(
            {
                "roc_auc_score": 1.0,
                "average_precision_score": 1.0,
                "roc_curve": [[0.0, 0.0, 0.0, 1.0], [0.0, 0.5, 1.0, 1.0]],
                "precision_recall_curve": [[0.6666666666666666, 1.0, 1.0, 1.0], [1.0, 1.0, 0.5, 0.0]],
                "calibration_curve": [[0.0, 1.0, 1.0], [0.1, 0.2, 0.9]],
                "calibration_error": 0.3333333333333333,
            }
        ),
    }

    assert evaluate_binary_classification(input) == expected_output


def test_evaluate_binary_classification_small_null_probabilities():
    input = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
            "predicted_boolean_value": [True, False, False],
            "predicted_boolean_probability": [None, None, None],
        },
        schema=BINARY_CLASSIFICATION_SCHEMA_DICT,
    )

    expected_output = {
        "samples_equally_weighted": pytest.approx(
            {
                "binary_accuracy": 0.6666666666666666,
                "f1_score": 0.6666666666666666,
            }
        ),
        "subjects_equally_weighted": pytest.approx(
            {
                "binary_accuracy": 0.6666666666666666,
                "f1_score": 0.6666666666666666,
            }
        ),
    }

    assert evaluate_binary_classification(input) == expected_output


# TODO add tests for schema validation



from datetime import datetime

import polars as pl
import pytest

from meds_evaluation.schema import BINARY_CLASSIFICATION_SCHEMA_DICT, validate_binary_classification_schema


def test_validate_binary_classification_schema():
    df = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
            "predicted_boolean_value": [True, False, False],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
        },
        schema=BINARY_CLASSIFICATION_SCHEMA_DICT,
    )

    df_reordered = pl.DataFrame(
        {
            "predicted_boolean_value": [True, False, False],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
            "subject_id": [1, 2, 3],
        },
        schema=BINARY_CLASSIFICATION_SCHEMA_DICT,
    )
    assert validate_binary_classification_schema(df) is None
    assert validate_binary_classification_schema(df_reordered) is None


def test_validate_binary_classification_schema_missing_required_field():
    df_no_subject_id = pl.DataFrame(
        {
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
            "predicted_boolean_value": [True, False, False],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
        },
    )

    df_no_prediction_time = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "boolean_value": [True, False, True],
            "predicted_boolean_value": [True, False, False],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
        },
    )

    df_no_boolean_value = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "predicted_boolean_value": [True, False, False],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
        },
    )

    with pytest.raises(ValueError, match="Missing required fields"):
        validate_binary_classification_schema(df_no_subject_id)

    with pytest.raises(ValueError, match="Missing required fields"):
        validate_binary_classification_schema(df_no_prediction_time)

    with pytest.raises(ValueError, match="Missing required fields"):
        validate_binary_classification_schema(df_no_boolean_value)


def test_validate_binary_classification_schema_missing_all_prediction_fields():
    df_no_predictions = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
        },
    )

    with pytest.raises(ValueError, match="Missing all prediction fields"):
        validate_binary_classification_schema(df_no_predictions)


def test_validate_binary_classification_schema_null_predictions():
    with pytest.raises(ValueError, match="At least one of the prediction fields should have non-null values"):
        df_null_values_no_probabilities = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_value": [None, None, None],
            },
        )
        validate_binary_classification_schema(df_null_values_no_probabilities)

    with pytest.raises(ValueError, match="At least one of the prediction fields should have non-null values"):
        df_no_values_null_probabilities = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_probability": [None, None, None],
            },
        )
        validate_binary_classification_schema(df_no_values_null_probabilities)

    with pytest.raises(ValueError, match="At least one of the prediction fields should have non-null values"):
        df_null_values_null_probabilities = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_value": [None, None, None],
                "predicted_boolean_probability": [None, None, None],
            },
            schema=BINARY_CLASSIFICATION_SCHEMA_DICT,
        )
        validate_binary_classification_schema(df_null_values_null_probabilities)


def test_validate_binary_classification_schema_mismatched_types():
    with pytest.raises(ValueError, match="Mismatched type for subject_id"):
        df_float_subject_id = pl.DataFrame(
            {
                "subject_id": [1.618, 2.718, 3.141],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_value": [True, False, False],
                "predicted_boolean_probability": [0.9, 0.1, 0.2],
            },
        )

        validate_binary_classification_schema(df_float_subject_id)

    with pytest.raises(ValueError, match="Mismatched type for prediction_time"):
        df_string_prediction_time = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    "2020-01-01 12:00:00",
                    "2020-01-01 12:00:00",
                    "2020-01-01 12:00:00",
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_value": [True, False, False],
                "predicted_boolean_probability": [0.9, 0.1, 0.2],
            },
        )
        validate_binary_classification_schema(df_string_prediction_time)

    with pytest.raises(ValueError, match="Mismatched type for boolean_value"):
        df_int_boolean_value = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [1, 0, 1],
                "predicted_boolean_value": [True, False, False],
                "predicted_boolean_probability": [0.9, 0.1, 0.2],
            },
        )
        validate_binary_classification_schema(df_int_boolean_value)

    with pytest.raises(ValueError, match="Mismatched type for predicted_boolean_value"):
        df_int_predicted_boolean_value = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_value": [1, 0, 0],
                "predicted_boolean_probability": [0.9, 0.1, 0.2],
            },
        )
        validate_binary_classification_schema(df_int_predicted_boolean_value)

    with pytest.raises(ValueError, match="Mismatched type for predicted_boolean_value"):
        df_float_predicted_boolean_value = pl.DataFrame(
            {
                "subject_id": [1, 2, 3],
                "prediction_time": [
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                    datetime(2020, 1, 1, 12, 0, 0),
                ],
                "boolean_value": [True, False, True],
                "predicted_boolean_value": [1.0, 0.0, 0.0],
                "predicted_boolean_probability": [0.9, 0.1, 0.2],
            },
        )
        validate_binary_classification_schema(df_float_predicted_boolean_value)


def test_validate_binary_classification_schema_extra_fields():
    df_with_embeddings = pl.DataFrame(
        {
            "subject_id": [1, 2, 3],
            "prediction_time": [
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
                datetime(2020, 1, 1, 12, 0, 0),
            ],
            "boolean_value": [True, False, True],
            "predicted_boolean_value": [True, False, False],
            "predicted_boolean_probability": [0.9, 0.1, 0.2],
            "embeddings": [1.618, 2.718, 3.141],
        },
    )

    assert validate_binary_classification_schema(df_with_embeddings) is None
